# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Romain Courteaud <romain@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

import re
import requests
import sys
import traceback
from urllib.parse import urlparse, urlunsplit
from forcediphttpsadapter.adapters import ForcedIPHTTPSAdapter
from peewee import fn


PREFERRED_TYPE = "text/html"
TIMEOUT = 2
ELAPSED_FAST = 0.2
ELAPSED_MODERATE = 0.5
SUB_CSP_NONCE_REPL = "'nonce-'"
SUB_CSP_NONCE = re.compile(r"'nonce-[^']*'").sub


def getUrlHostname(url):
    return urlparse(url).hostname


def getRootUrl(url):
    parsed_url = urlparse(url)
    return "%s://%s/" % (parsed_url.scheme, parsed_url.hostname)


def getUserAgent(version, contact=None):
    if contact:
        contact = " (+%s)" % contact
    else:
        contact = ""
    return "%s/%s%s" % ("SURYKATKA", version, contact)


def request(
    url,
    timeout=TIMEOUT,
    headers=None,
    session=requests,
    contact=None,
    version=0,
):

    if headers is None:
        headers = {}
    if "Accept" not in headers:
        headers["Accept"] = "%s;q=0.9,*/*;q=0.8" % PREFERRED_TYPE
    if "User-Agent" not in headers:
        # XXX user agent
        headers["User-Agent"] = getUserAgent(version, contact)

    kwargs = {}
    kwargs["stream"] = False
    kwargs["timeout"] = timeout
    kwargs["allow_redirects"] = False
    kwargs["verify"] = True
    args = ["GET", url]

    kwargs["headers"] = headers

    try:
        response = session.request(*args, **kwargs)
    except requests.exceptions.SSLError:
        # XXX Enter into unknown host
        response = requests.models.Response()
        response.status_code = 526
    except requests.exceptions.ConnectionError:
        response = requests.models.Response()
        response.status_code = 523
    except (
        requests.exceptions.Timeout,
        requests.exceptions.ChunkedEncodingError,
        requests.exceptions.ContentDecodingError,
    ):
        response = requests.models.Response()
        response.status_code = 524
    except requests.exceptions.TooManyRedirects:
        response = requests.models.Response()
        response.status_code = 520
    except requests.exceptions.RequestException:
        response = requests.models.Response()
        response.status_code = 000
        print(traceback.print_exception(*(sys.exc_info())))
    return response


def reportHttp(db, ip=None, url=None):
    query = (
        db.HttpCodeChange.select(db.HttpCodeChange)
        .group_by(db.HttpCodeChange.ip, db.HttpCodeChange.url)
        .having(
            db.HttpCodeChange.status_id == fn.MAX(db.HttpCodeChange.status_id)
        )
    )

    if ip is not None:
        if type(ip) == list:
            query = query.where(db.HttpCodeChange.ip << ip)
        else:
            query = query.where(db.HttpCodeChange.ip == ip)

    if url is not None:
        if type(url) == list:
            query = query.where(db.HttpCodeChange.url << url)
        else:
            query = query.where(db.HttpCodeChange.url == url)

    return query


def packHttp(db):
    with db._db.atomic():
        result = [x for x in reportHttp(db)]
        for http_change in result:
            db.HttpCodeChange.delete().where(
                db.HttpCodeChange.status_id != http_change.status_id,
                db.HttpCodeChange.url == http_change.url,
                db.HttpCodeChange.ip == http_change.ip,
            ).execute()


def calculateSpeedRange(total_seconds, fast, moderate):
    # Prevent updating the DB by defining acceptable speed range
    if total_seconds == 0:
        # error cases
        return "BAD"
    elif total_seconds < fast:
        return "FAST"
    elif total_seconds < moderate:
        return "MODERATE"
    else:
        return "SLOW"


def logHttpStatus(
    db,
    ip,
    url,
    code,
    http_header_dict,
    total_seconds,
    fast,
    moderate,
    status_id,
):

    with db._db.atomic():
        try:
            # Check previous parameter value
            previous_entry = reportHttp(db, ip=ip, url=url).get()
        except db.HttpCodeChange.DoesNotExist:
            previous_entry = None

        if (
            (previous_entry is None)
            or (previous_entry.status_code != code)
            or (previous_entry.http_header_dict != http_header_dict)
            or (
                calculateSpeedRange(
                    previous_entry.total_seconds, fast, moderate
                )
                != calculateSpeedRange(total_seconds, fast, moderate)
            )
        ):
            previous_entry = db.HttpCodeChange.create(
                status=status_id,
                ip=ip,
                url=url,
                status_code=code,
                http_header_dict=http_header_dict,
                total_seconds=total_seconds,
            )
        return previous_entry.status_id


def checkHttpStatus(
    db,
    status_id,
    url,
    ip,
    bot_version,
    contact,
    timeout=TIMEOUT,
    elapsed_fast=ELAPSED_FAST,
    elapsed_moderate=ELAPSED_MODERATE,
):
    parsed_url = urlparse(url)
    hostname = parsed_url.hostname
    request_kw = {"timeout": timeout}
    # SNI Support
    if parsed_url.scheme == "https":
        # Provide SNI support
        base_url = urlunsplit(
            (parsed_url.scheme, parsed_url.netloc, "", "", "")
        )
        session = requests.Session()
        session.mount(base_url, ForcedIPHTTPSAdapter(dest_ip=ip))
        request_kw["session"] = session
        ip_url = url
    elif parsed_url.scheme == "http":
        # Force IP location
        parsed_url = parsed_url._replace(netloc=ip)
        ip_url = parsed_url.geturl()
    else:
        raise NotImplementedError("Unhandled url: %s" % url)

    response = request(
        ip_url,
        headers={"Host": hostname},
        contact=contact,
        version=bot_version,
        **request_kw,
    )

    # Blacklisted, because of non stability
    # 'Date'
    header_list = [
        # Redirect
        "Location",
        # HTTP Range
        "Accept-Ranges",
        # HTTP Cache
        "Vary",
        "Cache-Control",
        "WWW-Authenticate",
        # gzip
        "Content-Type",
        "Content-Encoding",
        "Content-Disposition",
        # Security
        "Referrer-Policy",
        "Strict-Transport-Security",
        "Feature-Policy",
        "X-Frame-Options",
        "X-Content-Type-Options",
        # CORS
        "Access-Control-Allow-Origin",
        "Access-Control-Allow-Methods",
        "Access-Control-Allow-Credentials",
        "Access-Control-Allow-Headers",
        "Access-Control-Expose-Headers",
    ]
    header_dict = {}
    for header_key in header_list:
        header_value = response.headers.get(header_key, None)
        if header_value is not None:
            header_dict[header_key] = header_value

    # Strip nonces from Content-Security-Policy
    header_key = "Content-Security-Policy"
    content_security_policy = response.headers.get(header_key, None)
    if content_security_policy is not None:
        header_dict[header_key] = SUB_CSP_NONCE(
            repl=SUB_CSP_NONCE_REPL,
            string=content_security_policy,
        )

    # Store key only, because of non stability
    # 'Etag', 'Last-Modified', 'Set-Cookie', 'Date', 'Age'
    key_only_header_list = [
        "Etag",
        "Last-Modified",
        "Set-Cookie",
        "Age",
        # Expires is dropped by some CDN when getting a cache hit
        # "Expires",
    ]
    for header_key in key_only_header_list:
        header_value = response.headers.get(header_key, None)
        if header_value is not None:
            header_dict[header_key] = True

    logHttpStatus(
        db,
        ip,
        url,
        response.status_code,
        header_dict,
        response.elapsed.total_seconds(),
        elapsed_fast,
        elapsed_moderate,
        status_id,
    )



# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Romain Courteaud <romain@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

import miniupnpc
from dns.resolver import get_default_resolver
import platform
import socket


def checkPlatform(version):
    config = {
        "platform": platform.platform(),
        "python_build": platform.python_build(),
        "python_compiler": platform.python_compiler(),
        "python_branch": platform.python_branch(),
        "python_implementation": platform.python_implementation(),
        "python_revision": platform.python_revision(),
        "python_version": platform.python_version(),
        "hostname": socket.gethostname(),
        "version": version,
    }

    config["resolvers"] = get_default_resolver().nameservers

    u = miniupnpc.UPnP()
    u.discoverdelay = 1000
    u.discover()
    try:
        u.selectigd()
        config["ip"] = u.externalipaddress()
    except Exception:
        config["ip"] = None
    return config


def logPlatform(db, version, status_id):
    config = checkPlatform(version)
    with db._db.atomic():
        for key, value in config.items():
            value = str(value)
            try:
                # Check previous parameter value
                previous_value = (
                    db.PlatformChange.select()
                    .where(db.PlatformChange.parameter == key)
                    .order_by(db.PlatformChange.status.desc())
                    .get()
                    .value
                )
            except db.PlatformChange.DoesNotExist:
                previous_value = None

            if previous_value != value:
                db.PlatformChange.create(
                    status=status_id, parameter=key, value=value
                )



# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Romain Courteaud <romain@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

import click
import sys
from .bot import create_bot


@click.command(short_help="Runs surykatka bot.")
@click.option(
    "--run",
    "-r",
    help="The bot operation mode to run.",
    show_default=True,
    default="status",
    type=click.Choice(
        ["crawl", "pack", "status", "recent", "warning", "error"]
    ),
)
@click.option(
    "--sqlite", "-s", help="The path of the sqlite DB. (default: :memory:)"
)
@click.option("--nameserver", "-n", help="The IP of the DNS server.")
@click.option("--url", "-u", help="The url to check.")
@click.option("--domain", "-d", help="The domain to check.")
@click.option("--timeout", "-t", help="The timeout value (in second).")
@click.option("--warning", "-w", help="The warning duration value (in days).")
@click.option(
    "--configuration", "-f", help="The path of the configuration file."
)
@click.option(
    "--reload/--no-reload",
    default=False,
    help="Reload the configuration file between each crawl.",
    show_default=True,
)
@click.option(
    "--stdout", help="File to store status. (default: stdout)", default=""
)
@click.option(
    "--output",
    "-o",
    help="The status output format.",
    type=click.Choice(["plain", "json"]),
    default="plain",
    show_default=True,
)
@click.option(
    "--profile", help="Profiler data path", type=click.Path(exists=False)
)
def runSurykatka(
    run,
    sqlite,
    nameserver,
    url,
    domain,
    timeout,
    warning,
    configuration,
    reload,
    stdout,
    output,
    profile,
):

    mapping = {}
    if url:
        mapping["URL"] = url
        mapping["DOMAIN"] = ""
    if domain:
        mapping["DOMAIN"] = domain
        if not url:
            mapping["URL"] = ""
    if timeout:
        mapping["TIMEOUT"] = timeout
    if warning:
        mapping["WARNING_PERIOD"] = str(60 * 60 * 24 * float(warning))
    if sqlite:
        mapping["SQLITE"] = sqlite
    if nameserver:
        mapping["NAMESERVER"] = nameserver
    if reload:
        mapping["RELOAD"] = str(reload)
    mapping["STDOUT"] = stdout
    mapping["FORMAT"] = output
    bot = create_bot(cfgfile=configuration, mapping=mapping)
    if profile is None:
        return bot.run(run)
    else:
        import cProfile

        return cProfile.runctx(
            "bot.run(run)", globals(), locals(), filename=profile
        )


@click.command(short_help="Stats profiler bot data.")
@click.option("--stats", type=click.Choice(["cumul", "time"]))
@click.argument("profile", type=click.Path(exists=True, dir_okay=False))
def runStats(stats, profile):
    click.echo("Profile bot execution")
    import pstats

    profile_stats = pstats.Stats(profile)
    if stats == "time":
        profile_stats.sort_stats("time", "calls").print_stats(30)
    else:
        profile_stats.sort_stats("cumulative").print_stats(30)


if __name__ == "__main__":
    sys.exit(runSurykatka())



# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Romain Courteaud <romain@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

from peewee import fn
import socket
import ssl
import hashlib
from binascii import hexlify
import datetime


TIMEOUT = 2


def reportSslCertificate(db, ip=None, port=None, hostname=None):
    query = (
        db.SslChange.select(db.SslChange)
        .group_by(
            db.SslChange.ip,
            db.SslChange.port,
            db.SslChange.hostname,
        )
        .having(db.SslChange.status_id == fn.MAX(db.SslChange.status_id))
    )

    if hostname is not None:
        if type(hostname) == list:
            query = query.where(db.SslChange.hostname << hostname)
        else:
            query = query.where(db.SslChange.hostname == hostname)
    if port is not None:
        if type(port) == list:
            query = query.where(db.SslChange.port << port)
        else:
            query = query.where(db.SslChange.port == port)
    if ip is not None:
        if type(ip) == list:
            query = query.where(db.SslChange.ip << ip)
        else:
            query = query.where(db.SslChange.ip == ip)
    return query


def packSslCertificate(db):
    with db._db.atomic():
        result = [x for x in reportSslCertificate(db)]
        for ssl_change in result:
            db.SslChange.delete().where(
                db.SslChange.status_id != ssl_change.status_id,
                db.SslChange.hostname == ssl_change.hostname,
                db.SslChange.port == ssl_change.port,
                db.SslChange.ip == ssl_change.ip,
            ).execute()


def logSslCertificate(
    db,
    ip,
    port,
    hostname,
    sha1_fingerprint,
    not_before,
    not_after,
    subject,
    issuer,
    status_id,
):

    with db._db.atomic():
        try:
            # Check previous parameter value
            previous_entry = reportSslCertificate(
                db, ip=ip, port=port, hostname=hostname
            ).get()
        except db.SslChange.DoesNotExist:
            previous_entry = None

        if (previous_entry is None) or (
            previous_entry.sha1_fingerprint != sha1_fingerprint
        ):
            previous_entry = db.SslChange.create(
                status=status_id,
                ip=ip,
                port=port,
                hostname=hostname,
                sha1_fingerprint=sha1_fingerprint,
                not_before=not_before,
                not_after=not_after,
                subject=subject,
                issuer=issuer,
            )
        return previous_entry.status_id


def hasValidSSLCertificate(db, ip, port, hostname, status_id, timeout=TIMEOUT):
    ssl_context = ssl.create_default_context()
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(timeout)
    wrapped_sock = ssl_context.wrap_socket(sock, server_hostname=hostname)

    try:
        wrapped_sock.connect((ip, port))
        der = wrapped_sock.getpeercert(True)
        # XXX How to extract info from the der directly?
        ssl_info = wrapped_sock.getpeercert()

    except (ssl.SSLError, ConnectionRefusedError, socket.timeout, OSError):
        wrapped_sock.close()
        # XXX Expired certificate can not be fetched with the builtin ssl lib
        # pyOpenSSL is one way to fix this
        # https://stackoverflow.com/a/52298575
        logSslCertificate(
            db,
            ip,
            port,
            hostname,
            None,
            None,
            None,
            None,
            None,
            status_id,
        )
        return False
    except:
        wrapped_sock.close()
        raise

    wrapped_sock.close()

    sha1_fingerprint = hexlify(hashlib.sha1(der).digest())
    ssl_date_fmt = "%b %d %H:%M:%S %Y %Z"
    not_before = datetime.datetime.strptime(
        ssl_info["notBefore"], ssl_date_fmt
    )
    not_after = datetime.datetime.strptime(ssl_info["notAfter"], ssl_date_fmt)
    subject = dict([y for x in ssl_info["subject"] for y in x]).get(
        "commonName", ""
    )
    issuer = dict([y for x in ssl_info["issuer"] for y in x]).get(
        "commonName", ""
    )
    logSslCertificate(
        db,
        ip,
        port,
        hostname,
        sha1_fingerprint.decode(),
        not_before,
        not_after,
        subject,
        issuer,
        status_id,
    )
    return True



# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Romain Courteaud <romain@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

import configparser
import os
from dns.resolver import get_default_resolver


CONFIG_SECTION = "SURYKATKA"


def createConfiguration(
    envvar="SURYKATKA_SETTINGS", cfgfile=None, mapping=None
):
    config = configparser.ConfigParser(empty_lines_in_values=False)
    # Default values
    config[CONFIG_SECTION] = {"INTERVAL": -1, "DOMAIN": "", "URL": ""}

    # User defined values
    if (envvar is not None) and (envvar in os.environ):
        config.read([os.environ.get(envvar)])
    if cfgfile is not None:
        config.read([cfgfile])
    if mapping is not None:
        config.read_dict({CONFIG_SECTION: mapping})

    # Required values
    if "SQLITE" not in config[CONFIG_SECTION]:
        config[CONFIG_SECTION]["SQLITE"] = ":memory:"
    if "NAMESERVER" not in config[CONFIG_SECTION]:
        config[CONFIG_SECTION]["NAMESERVER"] = "\n".join(
            get_default_resolver().nameservers
        )

    if "PUBLIC_SUFFIX" not in config[CONFIG_SECTION]:
        config[CONFIG_SECTION]["PUBLIC_SUFFIX"] = ""
    if "DOMAIN" not in config[CONFIG_SECTION]:
        config[CONFIG_SECTION]["DOMAIN"] = ""
    if "URL" not in config[CONFIG_SECTION]:
        config[CONFIG_SECTION]["URL"] = ""
    if "FORMAT" not in config[CONFIG_SECTION]:
        config[CONFIG_SECTION]["FORMAT"] = "json"
    if "WARNING_PERIOD" not in config[CONFIG_SECTION]:
        config[CONFIG_SECTION]["WARNING_PERIOD"] = str(60 * 60 * 24 * 7)
    if "TIMEOUT" not in config[CONFIG_SECTION]:
        config[CONFIG_SECTION]["TIMEOUT"] = "1"
    if "ELAPSED_FAST" not in config[CONFIG_SECTION]:
        config[CONFIG_SECTION]["ELAPSED_FAST"] = "-1"
    if "ELAPSED_MODERATE" not in config[CONFIG_SECTION]:
        config[CONFIG_SECTION]["ELAPSED_MODERATE"] = "-1"
    if "CONTACT" not in config[CONFIG_SECTION]:
        config[CONFIG_SECTION][
            "CONTACT"
        ] = "https://lab.nexedi.com/nexedi/surykatka"
    if "RELOAD" not in config[CONFIG_SECTION]:
        config[CONFIG_SECTION]["RELOAD"] = str(False)

    if config[CONFIG_SECTION]["SQLITE"] == ":memory:":
        # Do not loop when using temporary DB
        config[CONFIG_SECTION]["INTERVAL"] = "-1"

    return config[CONFIG_SECTION]


def logConfiguration(db, status_id, config):
    with db._db.atomic():
        for key, value in config.items():
            try:
                # Check previous parameter value
                previous_value = (
                    db.ConfigurationChange.select()
                    .where(db.ConfigurationChange.parameter == key)
                    .order_by(db.ConfigurationChange.status.desc())
                    .get()
                    .value
                )
            except db.ConfigurationChange.DoesNotExist:
                previous_value = None

            if previous_value != value:
                db.ConfigurationChange.create(
                    status=status_id, parameter=key, value=value
                )



# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Romain Courteaud <romain@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

import peewee
from playhouse.sqlite_ext import SqliteExtDatabase, JSONField
import datetime
from playhouse.migrate import migrate, SqliteMigrator


class LogDB:
    def __init__(self, sqlite_path):
        self._db = SqliteExtDatabase(
            sqlite_path,
            pragmas=(
                ("journal_mode", "WAL"),
                ("foreign_keys", 1),
                ("wal_autocheckpoint", 5),
                ("temp_store", "MEMORY"),
                ("synchronous", "NORMAL"),
                ("mmap_size", 30000000000),
            ),
        )
        self._db.connect()

        class BaseModel(peewee.Model):
            class Meta:
                database = self._db

        # This store the start, stop, loop time of the bot
        # All other tables point to it to be able to group some info
        class Status(BaseModel):
            text = peewee.TextField()
            timestamp = peewee.TimestampField(
                primary_key=True,
                # Store millisecond resolution
                resolution=6,
                # date is in UTC
                utc=True,
                default=datetime.datetime.utcnow,
            )

        # Store the configuration modification
        class ConfigurationChange(BaseModel):
            status = peewee.ForeignKeyField(Status)
            parameter = peewee.TextField(index=True)
            value = peewee.TextField()

            class Meta:
                primary_key = peewee.CompositeKey("status", "parameter")
                # indexes = (
                # create a unique on from/to/date
                # (('status', 'parameter'), True),
                # )

        # Store the configuration modification
        class PlatformChange(BaseModel):
            status = peewee.ForeignKeyField(Status)
            parameter = peewee.TextField(index=True)
            value = peewee.TextField()

            class Meta:
                primary_key = peewee.CompositeKey("status", "parameter")

        # Store remote network status
        class NetworkChange(BaseModel):
            status = peewee.ForeignKeyField(Status)
            ip = peewee.TextField()
            transport = peewee.TextField()
            port = peewee.IntegerField()
            state = peewee.TextField()

            class Meta:
                primary_key = peewee.CompositeKey(
                    "status", "ip", "transport", "port"
                )
                indexes = ((("ip", "transport", "port", "status"), True),)

        class DnsChange(BaseModel):
            status = peewee.ForeignKeyField(Status)
            resolver_ip = peewee.TextField()
            domain = peewee.TextField()
            rdtype = peewee.TextField()
            response = peewee.TextField()

            class Meta:
                primary_key = peewee.CompositeKey(
                    "status", "resolver_ip", "domain", "rdtype"
                )
                indexes = (
                    (("resolver_ip", "domain", "rdtype", "status"), True),
                )

        class DomainChange(BaseModel):
            status = peewee.ForeignKeyField(Status)
            domain = peewee.TextField()
            registrar = peewee.TextField(null=True)
            whois_server = peewee.TextField(null=True)
            creation_date = peewee.DateTimeField(null=True)
            updated_date = peewee.DateTimeField(null=True)
            expiration_date = peewee.DateTimeField(null=True)
            name_servers = peewee.TextField(null=True)
            whois_status = peewee.TextField(null=True)
            emails = peewee.TextField(null=True)
            dnssec = peewee.TextField(null=True)
            name = peewee.TextField(null=True)
            org = peewee.TextField(null=True)
            address = peewee.TextField(null=True)
            city = peewee.TextField(null=True)
            state = peewee.TextField(null=True)
            zipcode = peewee.TextField(null=True)
            country = peewee.TextField(null=True)

            class Meta:
                primary_key = peewee.CompositeKey("status", "domain")
                indexes = ((("domain", "status"), True),)

        class SslChange(BaseModel):
            status = peewee.ForeignKeyField(Status)
            ip = peewee.TextField()
            port = peewee.IntegerField()
            hostname = peewee.TextField()
            sha1_fingerprint = peewee.TextField(null=True)
            not_before = peewee.TimestampField(null=True, utc=True)
            not_after = peewee.TimestampField(null=True, utc=True)
            subject = peewee.TextField(null=True)
            issuer = peewee.TextField(null=True)

            class Meta:
                primary_key = peewee.CompositeKey(
                    "status", "ip", "port", "hostname"
                )
                indexes = ((("ip", "port", "hostname", "status"), True),)

        class HttpCodeChange(BaseModel):
            status = peewee.ForeignKeyField(Status)
            ip = peewee.TextField()
            url = peewee.TextField()
            status_code = peewee.IntegerField()
            http_header_dict = JSONField(default=dict)
            total_seconds = peewee.FloatField(default=0)

            class Meta:
                primary_key = peewee.CompositeKey("status", "ip", "url")
                indexes = ((("ip", "url", "status"), True),)

        self.Status = Status
        self.ConfigurationChange = ConfigurationChange
        self.PlatformChange = PlatformChange
        self.NetworkChange = NetworkChange
        self.DnsChange = DnsChange
        self.DomainChange = DomainChange
        self.HttpCodeChange = HttpCodeChange
        self.SslChange = SslChange

    def createTables(self):
        # http://www.sqlite.org/pragma.html#pragma_user_version
        db_version = self._db.pragma("user_version")
        expected_version = 6
        if db_version != expected_version:
            with self._db.transaction():

                if db_version == 0:
                    # version 0 (no table)
                    self._db.create_tables(
                        [
                            self.Status,
                            self.ConfigurationChange,
                            self.HttpCodeChange,
                            self.NetworkChange,
                            self.PlatformChange,
                            self.DnsChange,
                        ]
                    )

                if db_version <= 1:
                    # version 1 without SSL support
                    self._db.create_tables([self.SslChange])

                if db_version <= 5:
                    self._db.create_tables([self.DomainChange])

                migrator = SqliteMigrator(self._db)
                migration_list = []

                if (0 < db_version) and (db_version <= 2):
                    # version 2 without the http total_seconds column
                    migration_list.append(
                        migrator.add_column(
                            "HttpCodeChange",
                            "total_seconds",
                            self.HttpCodeChange.total_seconds,
                        )
                    )

                if (0 < db_version) and (db_version <= 3):
                    # version 3 without the http header column
                    migration_list.append(
                        migrator.add_column(
                            "HttpCodeChange",
                            "http_header_dict",
                            self.HttpCodeChange.http_header_dict,
                        )
                    )

                if (0 < db_version) and (db_version <= 4):
                    # version 4 without the index to speed up reporting
                    migration_list.extend(
                        [
                            migrator.add_index(
                                "NetworkChange",
                                ("ip", "transport", "port", "status_id"),
                                True,
                            ),
                            migrator.add_index(
                                "DnsChange",
                                (
                                    "resolver_ip",
                                    "domain",
                                    "rdtype",
                                    "status_id",
                                ),
                                True,
                            ),
                            migrator.add_index(
                                "HttpCodeChange",
                                ("ip", "url", "status_id"),
                                True,
                            ),
                        ]
                    )
                if (1 < db_version) and (db_version <= 4):
                    # version 4 without the index to speed up reporting
                    migration_list.append(
                        migrator.add_index(
                            "SslChange",
                            ("ip", "port", "hostname", "status_id"),
                            True,
                        )
                    )

                if migration_list:
                    migrate(*migration_list)

                if db_version >= expected_version:
                    raise ValueError("Can not downgrade SQLite DB")

                self._db.pragma("user_version", expected_version)

    def close(self):
        self._db.close()

    def vacuum(self):
        self._db.execute_sql("VACUUM", [])



# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Romain Courteaud <romain@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

import io
import re
from setuptools import setup, find_packages

with io.open("src/surykatka/bot.py", "rt", encoding="utf8") as f:
    version = re.search(r'__version__ = "(.*?)"', f.read()).group(1)

setup(
    name="surykatka",
    version=version,
    license="GPLv3+",
    author="Nexedi",
    author_email="romain@nexedi.com",
    long_description=__doc__,
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=False,
    zip_safe=True,
    python_requires=">=3.5",
    install_requires=[
        "setuptools>40.5.0",
        "requests>2.20.0",
        "forcediphttpsadapter",
        "peewee>2.10.1",
        "click>=7.0",
        "dnspython",
        "python-whois",
        "miniupnpc",
    ],
    extras_require={
        "dev": ["pytest", "black", "pyflakes", "mock", "httpretty"]
    },
    entry_points={
        "console_scripts": [
            "surykatka=surykatka.cli:runSurykatka ",
            "surykatkastat=surykatka.cli:runStats ",
        ]
    },
)



# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Romain Courteaud <romain@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

import socket
import errno
from peewee import fn


TIMEOUT = 2


def reportNetwork(db, ip=None, transport=None, port=None):
    query = (
        db.NetworkChange.select(db.NetworkChange)
        .group_by(
            db.NetworkChange.ip,
            db.NetworkChange.transport,
            db.NetworkChange.port,
        )
        .having(
            db.NetworkChange.status_id == fn.MAX(db.NetworkChange.status_id)
        )
    )

    if transport is not None:
        if type(transport) == list:
            query = query.where(db.NetworkChange.transport << transport)
        else:
            query = query.where(db.NetworkChange.transport == transport)
    if port is not None:
        if type(port) == list:
            query = query.where(db.NetworkChange.port << port)
        else:
            query = query.where(db.NetworkChange.port == port)
    if ip is not None:
        if type(ip) == list:
            query = query.where(db.NetworkChange.ip << ip)
        else:
            query = query.where(db.NetworkChange.ip == ip)
    return query


def packNetwork(db):
    with db._db.atomic():
        result = [x for x in reportNetwork(db)]
        for network_change in result:
            db.NetworkChange.delete().where(
                db.NetworkChange.status_id != network_change.status_id,
                db.NetworkChange.transport == network_change.transport,
                db.NetworkChange.port == network_change.port,
                db.NetworkChange.ip == network_change.ip,
            ).execute()


def logNetwork(db, ip, transport, port, state, status_id):

    with db._db.atomic():
        try:
            # Check previous parameter value
            previous_entry = reportNetwork(
                db, ip=ip, transport=transport, port=port
            ).get()
        except db.NetworkChange.DoesNotExist:
            previous_entry = None

        if (previous_entry is None) or (previous_entry.state != state):
            previous_entry = db.NetworkChange.create(
                status=status_id,
                ip=ip,
                transport=transport,
                port=port,
                state=state,
            )
        return previous_entry.status_id


def isTcpPortOpen(db, ip, port, status_id, timeout=TIMEOUT):
    is_open = False
    sock = socket.socket()
    sock.settimeout(timeout)
    try:
        sock.connect((ip, port))
        state = "open"
        is_open = True
    except ConnectionRefusedError:
        state = "closed"
    except (socket.timeout, TimeoutError):
        state = "filtered"
    except OSError as e:
        if e.errno == errno.EHOSTUNREACH:
            # OSError: [Errno 113] No route to host
            state = "filtered"
        elif e.errno == errno.ENETUNREACH:
            # OSError: [Errno 101] Network is unreachable
            state = "unreachable"
        else:
            sock.close()
            raise
    except:
        sock.close()
        raise

    sock.close()

    logNetwork(db, ip, "TCP", port, state, status_id)
    return is_open



# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Romain Courteaud <romain@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

from dns import (
    resolver as dns_resolver,
    name as dns_name,
    exception as dns_exception,
    reversename as dns_reversename,
)
from .network import logNetwork
from peewee import fn

URL_TO_CHECK = "example.org"
TIMEOUT = 2


def reportDnsQuery(db, resolver_ip=None, domain=None, rdtype=None):
    query = (
        db.DnsChange.select(db.DnsChange)
        .group_by(
            db.DnsChange.resolver_ip, db.DnsChange.domain, db.DnsChange.rdtype
        )
        .having(db.DnsChange.status_id == fn.MAX(db.DnsChange.status_id))
    )

    if resolver_ip is not None:
        if type(resolver_ip) == list:
            query = query.where(db.DnsChange.resolver_ip << resolver_ip)
        else:
            query = query.where(db.DnsChange.resolver_ip == resolver_ip)
    if domain is not None:
        if type(domain) == list:
            query = query.where(db.DnsChange.domain << domain)
        else:
            query = query.where(db.DnsChange.domain == domain)
    if rdtype is not None:
        if type(rdtype) == list:
            query = query.where(db.DnsChange.rdtype << rdtype)
        else:
            query = query.where(db.DnsChange.rdtype == rdtype)
    return query


def packDns(db):
    with db._db.atomic():
        result = [x for x in reportDnsQuery(db)]
        for dns_change in result:
            db.DnsChange.delete().where(
                db.DnsChange.status_id != dns_change.status_id,
                db.DnsChange.resolver_ip == dns_change.resolver_ip,
                db.DnsChange.domain == dns_change.domain,
                db.DnsChange.rdtype == dns_change.rdtype,
            ).execute()


def logDnsQuery(db, status_id, resolver_ip, domain_text, rdtype, answer_list):
    answer_list.sort()
    response = ", ".join(answer_list)

    with db._db.atomic():
        try:
            # Check previous parameter value
            previous_entry = reportDnsQuery(
                db, resolver_ip=resolver_ip, domain=domain_text, rdtype=rdtype
            ).get()
        except db.DnsChange.DoesNotExist:
            previous_entry = None

        if (previous_entry is None) or (previous_entry.response != response):
            previous_entry = db.DnsChange.create(
                resolver_ip=resolver_ip,
                domain=domain_text,
                rdtype=rdtype,
                response=response,
                status=status_id,
            )

    return previous_entry.status_id


def buildResolver(resolver_ip, timeout):
    resolver = dns_resolver.Resolver(configure=False)
    resolver.nameservers.append(resolver_ip)
    resolver.timeout = timeout
    resolver.lifetime = timeout
    resolver.edns = -1
    return resolver


def reverseIp(ip):
    return dns_reversename.from_address(ip)


def queryDNS(db, status_id, resolver_ip, domain_text, rdtype, timeout=TIMEOUT):
    # only A (and AAAA) has address property
    assert rdtype in ["A", "MX", "TXT", "PTR"], rdtype

    resolver = buildResolver(resolver_ip, timeout)
    try:
        answer_list = [
            (
                x.address
                if (rdtype == "A")
                else (
                    x.exchange.derelativize(domain_text).to_text()[:-1]
                    if (rdtype == "MX")
                    else (
                        x.to_text()[:-1] if (rdtype == "PTR") else x.to_text()
                    )
                )
            )
            for x in resolver.query(
                domain_text, rdtype, raise_on_no_answer=False
            )
        ]
    except (
        dns_resolver.NXDOMAIN,
        dns_resolver.NoAnswer,
        dns_exception.Timeout,
        dns_resolver.NoNameservers,
    ):
        answer_list = []
        # how to differentiate no answer from empty answer

    logDnsQuery(db, status_id, resolver_ip, domain_text, rdtype, answer_list)
    return answer_list


def getReachableResolverList(db, status_id, resolver_ip_list, timeout=TIMEOUT):
    # Create a list of resolver object
    result_ip_list = []
    # Check the DNS server availability once
    # to prevent using it later if it is down
    for resolver_ip in resolver_ip_list:
        resolver_state = "open"
        answer_list = queryDNS(
            db, status_id, resolver_ip, URL_TO_CHECK, "A", timeout
        )

        if len(answer_list) == 0:
            # We expect a valid response
            # Drop the DNS server...
            resolver_state = "closed"
        else:
            resolver_state = "open"
            result_ip_list.append(resolver_ip)
        logNetwork(db, resolver_ip, "UDP", 53, resolver_state, status_id)

    return result_ip_list


def expandDomainList(domain_list, public_suffix_list=None):
    for domain_text in domain_list:

        dns_name_value = dns_name.from_text(domain_text)
        if (len(dns_name_value.labels) - 1) > 2:
            # https://publicsuffix.org/list/public_suffix_list.dat
            parent_domain_text = dns_name_value.parent().to_text(
                omit_final_dot=True
            )
            if (public_suffix_list is None) or (
                parent_domain_text not in public_suffix_list
            ):
                domain_list.append(parent_domain_text)

    domain_list = list(set(domain_list))
    domain_list.sort()
    return domain_list


def getDomainIpDict(
    db, status_id, resolver_ip_list, domain_list, rdtype, timeout=TIMEOUT
):
    server_ip_dict = {}
    for domain_text in domain_list:
        for resolver_ip in resolver_ip_list:
            answer_list = queryDNS(
                db, status_id, resolver_ip, domain_text, rdtype, timeout
            )
            for address in answer_list:
                if address not in server_ip_dict:
                    server_ip_dict[address] = []
                if domain_text not in server_ip_dict[address]:
                    # Do not duplicate the domain
                    server_ip_dict[address].append(domain_text)
    return server_ip_dict



# Copyright (C) 2021  Nexedi SA and Contributors.
#                     Romain Courteaud <romain@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

import sys
import os
import whois
from peewee import fn


def reportWhoisQuery(db, domain=None):
    query = (
        db.DomainChange.select(db.DomainChange)
        .group_by(db.DomainChange.domain)
        .having(db.DomainChange.status_id == fn.MAX(db.DomainChange.status_id))
    )

    if domain is not None:
        if type(domain) == list:
            query = query.where(db.DomainChange.domain << domain)
        else:
            query = query.where(db.DomainChange.domain == domain)
    return query


def packDomain(db):
    with db._db.atomic():
        result = [x for x in reportWhoisQuery(db)]
        for dns_change in result:
            db.DomainChange.delete().where(
                db.DomainChange.status_id != dns_change.status_id,
                db.DomainChange.domain == dns_change.domain,
            ).execute()


def logWhoisQuery(
    db,
    status_id,
    domain_text,
    registrar,
    whois_server,
    creation_date,
    updated_date,
    expiration_date,
    name_servers,
    whois_status,
    emails,
    dnssec,
    name,
    org,
    address,
    city,
    state,
    zipcode,
    country,
):

    with db._db.atomic():
        try:
            # Check previous parameter value
            previous_entry = reportWhoisQuery(db, domain=domain_text).get()
        except db.DomainChange.DoesNotExist:
            previous_entry = None

        if (
            (previous_entry is None)
            or (previous_entry.registrar != registrar)
            or (previous_entry.whois_server != whois_server)
            or (previous_entry.creation_date != creation_date)
            or (previous_entry.updated_date != updated_date)
            or (previous_entry.expiration_date != expiration_date)
            or (previous_entry.name_servers != name_servers)
            or (previous_entry.whois_status != whois_status)
            or (previous_entry.emails != emails)
            or (previous_entry.dnssec != dnssec)
            or (previous_entry.name != name)
            or (previous_entry.org != org)
            or (previous_entry.address != address)
            or (previous_entry.city != city)
            or (previous_entry.state != state)
            or (previous_entry.zipcode != zipcode)
            or (previous_entry.country != country)
        ):
            previous_entry = db.DomainChange.create(
                domain=domain_text,
                registrar=registrar,
                whois_server=whois_server,
                creation_date=creation_date,
                updated_date=updated_date,
                expiration_date=expiration_date,
                name_servers=name_servers,
                whois_status=whois_status,
                emails=emails,
                dnssec=dnssec,
                name=name,
                org=org,
                address=address,
                city=city,
                state=state,
                zipcode=zipcode,
                country=country,
                status=status_id,
            )

    return previous_entry.status_id


def queryWhois(db, status_id, domain_text):
    # Hide lib message:
    # Error trying to connect to socket: closing socket
    _stdout = sys.stdout
    sys.stdout = open(os.devnull, "w")
    whois_dict = whois.whois(domain_text)
    sys.stdout = _stdout

    arg_list = []
    for arg in [
        whois_dict.registrar,
        whois_dict.whois_server,
        whois_dict.creation_date,
        whois_dict.updated_date,
        whois_dict.expiration_date,
        whois_dict.name_servers,
        whois_dict.status,
        whois_dict.emails,
        whois_dict.dnssec,
        whois_dict.name,
        whois_dict.org,
        whois_dict.address,
        whois_dict.city,
        whois_dict.state,
        whois_dict.zipcode,
        whois_dict.country,
    ]:
        if type(arg) == list:
            arg = arg[0]
        arg_list.append(arg)

    logWhoisQuery(db, status_id, domain_text, *arg_list)
    return whois_dict
